# Generated by Django 4.0.6 on 2022-08-17 15:26

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('idp_user', '0003_auto_20220513_1025'),
    ]

    operations = [
        migrations.RenameField(
            model_name='user',
            old_name='idp_user_id',
            new_name='id',
        ),
        migrations.AlterField(
            model_name='user',
            name='id',
            field=models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID'),
        ),
    ]
